/*
      Copyright (C) 2010 Bepu Entertainment LLC.

      This software source code is provided 'as-is', without 
      any express or implied warranty.  In no event will the authors be held 
      liable for any damages arising from the use of this software.

      Permission is granted to anyone to use this software for any purpose,
      including commercial applications, and to alter it and redistribute it
      freely, subject to the following restrictions:

      1. The origin of this software must not be misrepresented; you must not
         claim that you wrote the original software. If you use this software
         in a product, an acknowledgment in the product documentation would be
         appreciated but is not required.
      2. Altered source versions must be plainly marked as such, and must not be
         misrepresented as being the original software.
      3. This notice may not be removed or altered from any source distribution.

    Contact us at:
    contact@bepu-games.com
 */

#region Using Statements

using System;
using System.Collections.Generic;
using BEPUphysics;
using BEPUphysics.Entities;
using BEPUphysics.ResourceManagement;
using BEPUphysicsDemos.Demos;
using BEPUphysicsDemos.SampleCode;
using BEPUphysicsDrawer.Font;
using BEPUphysicsDrawer.Lines;
using BEPUphysicsDrawer.Models;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;

#endregion

namespace BEPUphysicsDemos
{
    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class DemosGame : Game
    {
        public Camera Camera;

        private int currentSimulationIndex = 1;
        private Demo currentSimulation;

        //Rendering Variables
        public GraphicsDeviceManager Graphics;
        private readonly List<Contact> contacts = new List<Contact>();
        private readonly List<VertexPositionColor> collisionNormalList = new List<VertexPositionColor>();
        private readonly List<VertexPositionColor> boundingBoxLines = new List<VertexPositionColor>();

        //Rendering tools
        public ModelDrawer ModelDrawer;
        public LineDrawer ConstraintDrawer;
        public BasicEffect LineDrawer;
        public SpriteBatch UIDrawer;
        public TextDrawer DataTextDrawer;
        public TextDrawer TinyTextDrawer;

        //Content
        private SpriteFont dataFont;
        private SpriteFont tinyFont;
        private Texture2D controlsMenu;

        //FPS calculation variables
        private double FPSlastTime;
        private double FPStotalSinceLast;
        private double FPStoDisplay;
        private double averagePhysicsTime;
        private int FPStotalFramesSinceLast;

        //Input
        public KeyboardState KeyboardInput;
        public KeyboardState PreviousKeyboardInput;
        public GamePadState GamePadInput;
        public GamePadState PreviousGamePadInput;
#if WINDOWS
        public MouseState MouseInput;
#endif

        //Display Booleans        
        private bool displayEntities = true;
        private bool displayUI = true;
        private bool displayConstraints = true;
        private bool displayMenu;
        private bool displayContacts;
        private bool displayBoundingBoxes;
        private bool displaySimulationIslands;

        private readonly Type[] demoTypes = new[]
                                                {
                                                    typeof (WallDemo),
                                                    typeof (IncomingDemo),
                                                    typeof (VariousThingsDemo),
                                                    typeof (TerrainDemo),
                                                    typeof (LotsOSpheresDemo),
                                                    typeof (MultipendulumDemo),
                                                    typeof (SawContraptionDemo),
                                                    typeof (PlinkoDemo),
                                                    typeof (DogbotDemo),
                                                    typeof (ActionFigureDemo),
                                                    typeof (CoefficientsDemo),
                                                    typeof (ColosseumDemo),
                                                    typeof (CollisionFilteringDemo),
                                                    typeof (UnfortunateGuyDemo),
                                                    typeof (RobotArmDemo),
                                                    typeof (PyramidDemo),
                                                    typeof (CompoundBodiesDemo),
                                                    typeof (SpaceshipDemo),
                                                    typeof (JengaDemo),
                                                    typeof (SleepModeDemo),
                                                    typeof (BridgeDemo),
                                                    typeof (ClothDemo),
                                                    typeof (BroadPhaseDemo),
                                                    typeof (PathFollowingDemo),
                                                    typeof (BuoyancyDemo),
                                                    typeof (TornadoDemo),
                                                    typeof (GravitationalFieldDemo),
                                                    typeof (PlaygroundDemo),
                                                    typeof (FancyShapesDemo),
                                                    typeof (FishInABarrelDemo),
                                                    typeof (EarthquakeDemo),
                                                    typeof (DetectorVolumeDemo),
                                                };


        public DemosGame()
        {
            Graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";

            Graphics.PreferredBackBufferWidth = 800;
            Graphics.PreferredBackBufferHeight = 600;
            Camera = new Camera(Vector3.Zero, 10, 0, 0, Matrix.CreatePerspectiveFieldOfView(MathHelper.PiOver4, 800 / 600f, .1f, 10000));


            Exiting += DemosGameExiting;
        }

        private void DemosGameExiting(object sender, EventArgs e)
        {
            currentSimulation.CleanUp();
        }


        protected override void Initialize()
        {
            if (GraphicsDevice.GraphicsProfile == GraphicsProfile.HiDef)
                ModelDrawer = new InstancedModelDrawer(this);
            else
                ModelDrawer = new BruteModelDrawer(this);

            ConstraintDrawer = new LineDrawer(this);
            ConstraintDrawer.DisplayTypes.Add(typeof(GrabSpring), typeof(DisplayGrabSpring));

            base.Initialize();
        }


        /// <summary>
        /// Manages the switch to a new physics engine simulation.
        /// </summary>
        /// <param name="sim">Index of the simulation to switch to.</param>
        private void SwitchSimulation(int sim)
        {
            currentSimulationIndex = sim;

            //Clear out any old rendering stuff.
            ModelDrawer.Clear();
            ConstraintDrawer.Clear();

            //Tell the previous simulation it's done.
            if (currentSimulation != null)
            {
                currentSimulation.CleanUp();
            }
            //Create the new demo.
            Type demoType = demoTypes[currentSimulationIndex - 1];

#if !WINDOWS
            currentSimulation = (Demo)demoType.GetConstructor(
                                                    new Type[] { typeof(DemosGame) })
                                                    .Invoke(new object[] { this });
#else
            currentSimulation = (Demo)Activator.CreateInstance(demoType, new object[] { this });
#endif

            #region DisplayObject creation

            foreach (Entity e in currentSimulation.Space.Entities)
            {
                if ((string)e.Tag != "noDisplayObject")
                {
                    ModelDrawer.Add(e);
                }
                else //Remove the now unnecessary tag.
                    e.Tag = null;
            }
            foreach (SolverUpdateable s in currentSimulation.Space.SolverUpdateables)
            {
                //Add the solver updateable and match up the activity setting.
                LineDisplayObjectBase objectAdded = ConstraintDrawer.Add(s);
                if (objectAdded != null)
                    objectAdded.IsActive = s.IsActive;
            }

            #endregion

            GC.Collect();
        }

        protected override void LoadContent()
        {
            dataFont = Content.Load<SpriteFont>("DataFont");
            tinyFont = Content.Load<SpriteFont>("TinyFont");

            controlsMenu = Content.Load<Texture2D>("bepuphysicscontrols");

            IsFixedTimeStep = false;

            LineDrawer = new BasicEffect(GraphicsDevice);

            UIDrawer = new SpriteBatch(GraphicsDevice);

            DataTextDrawer = new TextDrawer(UIDrawer, dataFont, Color.White);
            TinyTextDrawer = new TextDrawer(UIDrawer, tinyFont, Color.White);

#if WINDOWS
            Mouse.SetPosition(200, 200); //This helps the camera stay on track even if the mouse is offset during startup.
#endif

            SwitchSimulation(1);
        }


        protected override void UnloadContent()
        {
            base.UnloadContent();
        }

        /// <summary>
        /// Determines whether or not the key was pressed this frame.
        /// </summary>
        /// <param name="key">Key to check.</param>
        /// <returns>Whether or not the key was pressed.</returns>
        public bool WasKeyPressed(Keys key)
        {
            return KeyboardInput.IsKeyDown(key) && PreviousKeyboardInput.IsKeyUp(key);
        }

        /// <summary>
        /// Determines whether or not the button was pressed this frame.
        /// </summary>
        /// <param name="button">Button to check.</param>
        /// <returns>Whether or not the button was pressed.</returns>
        public bool WasButtonPressed(Buttons button)
        {
            return GamePadInput.IsButtonDown(button) && PreviousGamePadInput.IsButtonUp(button);
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            PreviousKeyboardInput = KeyboardInput;
            KeyboardInput = Keyboard.GetState();
            var dt = (float)gameTime.ElapsedGameTime.TotalSeconds;
#if WINDOWS
            MouseInput = Mouse.GetState();

            //Keep the mouse within the screen
            Mouse.SetPosition(200, 200);
#endif
            PreviousGamePadInput = GamePadInput;
            for (int i = 0; i < 4; i++)
            {
                GamePadInput = GamePad.GetState((PlayerIndex)i);
                if (GamePadInput.IsConnected)
                    break;
            }

            // Allows the default game to exit on Xbox 360 and Windows
            if (KeyboardInput.IsKeyDown(Keys.Escape) || GamePadInput.Buttons.Back == ButtonState.Pressed)
                Exit();

            #region Camera

            //Update the camera
#if !WINDOWS

            camera.update(dt, keyboardInput, gamePadInput);
#else
            Camera.Update(dt, KeyboardInput, MouseInput, GamePadInput);
#endif

            #endregion

            #region UI Toggles

#if !WINDOWS
            if (wasButtonPressed(Buttons.Start))
            {
                displayMenu = !displayMenu;
            }
#else
            if (WasKeyPressed(Keys.F1))
            {
                displayMenu = !displayMenu;
            }
#endif
            if (WasKeyPressed(Keys.I))
            {
                displayUI = !displayUI;
            }
            if (WasKeyPressed(Keys.J))
            {
                displayConstraints = !displayConstraints;
            }
            if (WasKeyPressed(Keys.K))
            {
                displayContacts = !displayContacts;
            }
            if (WasKeyPressed(Keys.U))
            {
                displayBoundingBoxes = !displayBoundingBoxes;
            }
            if (WasKeyPressed(Keys.Y))
            {
                displayEntities = !displayEntities;
            }
            if (WasKeyPressed(Keys.H))
            {
                displaySimulationIslands = !displaySimulationIslands;
            }
            if (WasKeyPressed(Keys.G))
            {
                ModelDrawer.IsWireframe = !ModelDrawer.IsWireframe;
            }

            #endregion

            #region Simulation Switcharoo

#if !WINDOWS

            int switchTo = -2;
            if (wasButtonPressed(Buttons.DPadLeft))
                switchTo = currentSimulationIndex - 1;
            else if (wasButtonPressed(Buttons.DPadRight))
                switchTo = currentSimulationIndex + 1;
            if (switchTo != -2)
            {
                if (switchTo < 1)
                    switchTo += demoTypes.Length;
                else if (switchTo > demoTypes.Length)
                    switchTo = 1;
                switchSimulation(switchTo);
            }
#else

            foreach (Keys key in KeyboardInput.GetPressedKeys())
            {
                int code = key.GetHashCode();

                if (code >= 48 && code <= 57)
                {
                    int simNumber;
                    if (code == 48)
                        simNumber = 10;
                    else
                        simNumber = code - 48;
                    if (KeyboardInput.IsKeyDown(Keys.LeftShift))
                        simNumber += 10;
                    if (KeyboardInput.IsKeyDown(Keys.LeftControl))
                        simNumber += 20;

                    if (simNumber <= demoTypes.Length)
                    {
                        (currentSimulation.Space.ThreadManager as IDisposable).Dispose(); //While it would clean up by itself, this gets it done a little quicker.
                        SwitchSimulation(simNumber);
                    }
                }
            }


#endif

            #endregion

            currentSimulation.Update(dt);

            ConstraintDrawer.Update();

            ModelDrawer.Update();
            base.Update(gameTime);
        }

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(new Color(.41f, .41f, .45f, 1));

            if (displayEntities)
                ModelDrawer.Draw(Camera.ViewMatrix, Camera.ProjectionMatrix);

            if (displayConstraints)
                ConstraintDrawer.Draw(Camera.ViewMatrix, Camera.ProjectionMatrix);

            LineDrawer.LightingEnabled = false;
            LineDrawer.VertexColorEnabled = true;
            LineDrawer.World = Matrix.Identity;
            LineDrawer.View = Camera.ViewMatrix;
            LineDrawer.Projection = Camera.ProjectionMatrix;

            #region Contact Drawing

            //Draw collisions.
            if (displayContacts)
            {
                contacts.Clear();
                foreach (CollisionPair pair in currentSimulation.Space.CollisionPairs)
                    contacts.AddRange(pair.Contacts);
                if (contacts.Count > 0)
                {
                    collisionNormalList.Clear();
                    for (int i = 0; i < contacts.Count; i++)
                    {
                        collisionNormalList.Add(new VertexPositionColor(contacts[i].Position, Color.White));
                        collisionNormalList.Add(new VertexPositionColor(contacts[i].Position + contacts[i].Normal, Color.White));
                    }

                    foreach (EffectPass pass in LineDrawer.CurrentTechnique.Passes)
                    {
                        pass.Apply();

                        GraphicsDevice.DrawUserPrimitives(PrimitiveType.LineList, collisionNormalList.ToArray(), 0, contacts.Count);
                    }
                }
            }

            #endregion

            #region bounding box drawing

            if (displayBoundingBoxes)
            {
                if (currentSimulation.Space.Entities.Count > 0)
                {
                    boundingBoxLines.Clear();

                    foreach (Entity e in currentSimulation.Space.Entities)
                    {
                        Vector3[] boundingBoxCorners = e.BoundingBox.GetCorners();
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[0], Color.DarkRed));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[1], Color.DarkRed));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[0], Color.DarkRed));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[3], Color.DarkRed));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[0], Color.DarkRed));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[4], Color.DarkRed));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[1], Color.DarkRed));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[2], Color.DarkRed));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[1], Color.DarkRed));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[5], Color.DarkRed));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[2], Color.DarkRed));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[3], Color.DarkRed));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[2], Color.DarkRed));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[6], Color.DarkRed));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[3], Color.DarkRed));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[7], Color.DarkRed));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[4], Color.DarkRed));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[5], Color.DarkRed));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[4], Color.DarkRed));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[7], Color.DarkRed));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[5], Color.DarkRed));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[6], Color.DarkRed));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[6], Color.DarkRed));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[7], Color.DarkRed));
                    }
                    foreach (EffectPass pass in LineDrawer.CurrentTechnique.Passes)
                    {
                        pass.Apply();
                        GraphicsDevice.DrawUserPrimitives(PrimitiveType.LineList, boundingBoxLines.ToArray(), 0, currentSimulation.Space.Entities.Count * 12);
                    }
                }
            }

            #endregion

            #region Simulation Island Drawing

            if (displaySimulationIslands)
            {
                if (currentSimulation.Space.Entities.Count > 0)
                {
                    boundingBoxLines.Clear();
                    BoundingBox box;

                    foreach (SimulationIsland s in currentSimulation.Space.SimulationIslands)
                    {
                        List<Entity> entityList = Resources.GetEntityList();
#if !WINDOWS
                        entityList.AddRange(s.Entities.Keys);
#else
                        entityList.AddRange(s.Entities);
#endif

                        box = entityList[0].BoundingBox;
                        for (int k = 1; k < entityList.Count; k++)
                        {
                            box = BoundingBox.CreateMerged(box, entityList[k].BoundingBox);
                        }
                        Resources.GiveBack(entityList);

                        Color colorToUse = s.IsCurrentlyActive ? Color.DarkGoldenrod : Color.DarkGray;
                        Vector3[] boundingBoxCorners = box.GetCorners();
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[0], colorToUse));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[1], colorToUse));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[0], colorToUse));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[3], colorToUse));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[0], colorToUse));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[4], colorToUse));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[1], colorToUse));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[2], colorToUse));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[1], colorToUse));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[5], colorToUse));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[2], colorToUse));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[3], colorToUse));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[2], colorToUse));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[6], colorToUse));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[3], colorToUse));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[7], colorToUse));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[4], colorToUse));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[5], colorToUse));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[4], colorToUse));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[7], colorToUse));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[5], colorToUse));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[6], colorToUse));

                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[6], colorToUse));
                        boundingBoxLines.Add(new VertexPositionColor(boundingBoxCorners[7], colorToUse));
                    }

                    if (currentSimulation.Space.SimulationIslands.Count > 0)
                        foreach (EffectPass pass in LineDrawer.CurrentTechnique.Passes)
                        {
                            pass.Apply();
                            GraphicsDevice.DrawUserPrimitives(PrimitiveType.LineList, boundingBoxLines.ToArray(), 0, currentSimulation.Space.SimulationIslands.Count * 12);
                        }
                }
            }

            #endregion

            #region UI Drawing

            UIDrawer.Begin();
            if (displayUI)
            {
                FPStotalSinceLast += gameTime.ElapsedGameTime.TotalSeconds;
                FPStotalFramesSinceLast++;
                if (gameTime.TotalGameTime.TotalSeconds - FPSlastTime > .25 && gameTime.ElapsedGameTime.TotalSeconds > 0)
                {
                    double avg = FPStotalSinceLast / FPStotalFramesSinceLast;
                    FPSlastTime = gameTime.TotalGameTime.TotalSeconds;
                    FPStoDisplay = Math.Round(1 / avg, 1);
                    averagePhysicsTime = Math.Round(1000 * currentSimulation.PhysicsTime, 1);
                    FPStotalSinceLast = 0;
                    FPStotalFramesSinceLast = 0;
                }

                DataTextDrawer.Draw("FPS: ", FPStoDisplay, new Vector2(50, 450));
                DataTextDrawer.Draw("Physics Time (ms): ", averagePhysicsTime, new Vector2(50, 467));
                DataTextDrawer.Draw("Collision Pairs: ", currentSimulation.Space.CollisionPairs.Count, new Vector2(50, 484));
                int countActive = 0;

                for (int i = 0; i < currentSimulation.Space.Entities.Count; i++)
                {
                    if (currentSimulation.Space.Entities[i].IsActive)
                        countActive++;
                }
                DataTextDrawer.Draw("Active Objects: ", countActive, new Vector2(50, 501));
#if !WINDOWS
                dataTextDrawer.draw("Press Start for Controls", new Vector2(50, 518));
#else
                DataTextDrawer.Draw("Press F1 for Controls", new Vector2(50, 518));
#endif

                TinyTextDrawer.Draw("Current Simulation: ", currentSimulationIndex, new Vector2(600, 500));
                TinyTextDrawer.Draw(currentSimulation.Name, new Vector2(620, 514));

                currentSimulation.DrawUI();
            }
            if (displayMenu)
            {
                UIDrawer.Draw(controlsMenu, new Rectangle(0, 0, 800, 600), Color.White);
            }
            UIDrawer.End();

            #endregion

            base.Draw(gameTime);
        }
    }
}